/**
  * Copyright 2011 National ICT Australia Limited
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
  *
  * http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  */
package com.nicta.scoobi.io.avro

import java.util.{Map => JMap}
import org.apache.commons.logging.LogFactory
import org.apache.hadoop.fs.Path
import org.apache.hadoop.mapred.FileAlreadyExistsException
import org.apache.hadoop.io.NullWritable
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat
import org.apache.hadoop.mapreduce.Job
import org.apache.avro.mapred.AvroKey
import org.apache.avro.mapreduce.AvroKeyOutputFormat

import com.nicta.scoobi.DList
import com.nicta.scoobi.DListPersister
import com.nicta.scoobi.io.DataSink
import com.nicta.scoobi.io.OutputConverter
import com.nicta.scoobi.io.Helper
import com.nicta.scoobi.impl.Configured


/** Smart functions for persisting distributed lists by storing them as Avro files. */
object AvroOutput {
  lazy val logger = LogFactory.getLog("scoobi.AvroOutput")


  /** Specify a distributed list to be persistent by storing it to disk as an Avro File. */
  def toAvroFile[B : AvroSchema](dl: DList[B], path: String, overwrite: Boolean = false): DListPersister[B] = {
    val sch = implicitly[AvroSchema[B]]

    val converter = new OutputConverter[AvroKey[sch.AvroType], NullWritable, B] {
      def toKeyValue(x: B) = (new AvroKey(sch.toAvro(x)), NullWritable.get)
    }

    val sink = new DataSink[AvroKey[sch.AvroType], NullWritable, B] with Configured {
      protected val outputPath = new Path(path)

      val outputFormat = classOf[AvroKeyOutputFormat[sch.AvroType]]
      val outputKeyClass = classOf[AvroKey[sch.AvroType]]
      val outputValueClass = classOf[NullWritable]

      def outputCheck {}

      def outputConfigure(job: Job) {
        configure(job)
        FileOutputFormat.setOutputPath(job, outputPath)
        job.getConfiguration.set("avro.schema.output.key", sch.schema.toString)
      }

      protected def checkPaths = {
        if (Helper.pathExists(outputPath)) {
          if (overwrite) {
            logger.info("Deleting the pre-existing output path: " + outputPath.toUri.toASCIIString)
            Helper.deletePath(outputPath)
          } else {
            throw new FileAlreadyExistsException("Output path already exists: " + outputPath)
          }
        } else {
          logger.info("Output path: " + outputPath.toUri.toASCIIString)
          logger.debug("Output Schema: " + sch.schema)
        }
      }

      lazy val outputConverter = converter
    }

    new DListPersister(dl, sink)
  }
}
